'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _fsExtra = require('fs-extra');

var _nodeElmCompiler = require('node-elm-compiler');

var _shelljs = require('shelljs');

var _shelljs2 = _interopRequireDefault(_shelljs);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _defaultOptions = require('./defaultOptions');

var _defaultOptions2 = _interopRequireDefault(_defaultOptions);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

const noop = () => {};

exports.default = (() => {
  var _ref = _asyncToGenerator(function* (userModuleFileName, {
    outputName = _defaultOptions2.default.outputName,
    projectDir,
    report = _defaultOptions2.default.report,
    pathToElmMake,
    argsToOutput = []
  } = {}) {
    // extract key paths
    const moduleId = `${+new Date()}${Math.round(Math.random() * 100000000)}`;
    const mainModule = `RunElmMain${moduleId}`;
    const templatePath = _path2.default.resolve(__dirname, 'RunElmMain.elm.template');
    const templateWithArgsPath = _path2.default.resolve(__dirname, 'RunElmMainWithArgs.elm.template');
    const userModulePath = _path2.default.resolve(userModuleFileName);
    const userModule = _path2.default.basename(userModulePath, '.elm');
    const resolvedProjectDir = projectDir ? _path2.default.resolve(projectDir) : _path2.default.resolve(_path2.default.dirname(userModulePath));
    const mainModuleFilename = _path2.default.join(resolvedProjectDir, `${mainModule}.elm`);
    const outputCompiledFilename = _path2.default.join(resolvedProjectDir, `run-elm-main-${moduleId}.js`);

    const elmCompileStdout = _path2.default.join(resolvedProjectDir, `stdout-${moduleId}.txt`);
    const elmCompileStderr = _path2.default.join(resolvedProjectDir, `stderr-${moduleId}.txt`);
    let elmCompileStdoutFd;
    let elmCompileStderrFd;

    try {
      // ensure global elm is installed (unless pathToElmMake is given)
      if (!pathToElmMake && !_shelljs2.default.which('elm')) {
        throw new Error('No elm global binary available. Please install with `npm install -g elm`.');
      }

      // ensure --output-name is specified adequately
      if (!outputName.match(/^[a-z_]\w*$/)) {
        throw new Error(`Provided output name \`${outputName}\` is not a valid constant or function name in elm.`);
      }
      if (['init', 'main', 'program', 'sendOutput'].includes(outputName)) {
        throw new Error(`It is not allowed to use \`${outputName}\` as a value for output name. Please rename the symbol you would like to output.`);
      }

      // ensure user module path is adequate
      try {
        const userModuleFileStats = yield (0, _fsExtra.stat)(userModulePath);
        if (!userModuleFileStats.isFile()) {
          throw new Error();
        }
      } catch (err) {
        throw new Error(`File \`${userModulePath}\` does not exist.`);
      }
      if (!userModulePath.match(/\.elm$/i, '')) {
        throw new Error(`File \`${userModulePath}\` should have .elm file extension.`);
      }

      // ensure project folder is adequate
      try {
        const projectDirStats = yield (0, _fsExtra.stat)(resolvedProjectDir);
        if (!projectDirStats.isDirectory()) {
          throw new Error();
        }
      } catch (err) {
        throw new Error(`Provided project dir \`${resolvedProjectDir}\` is not a directory.`);
      }
      if (!userModulePath.startsWith(`${resolvedProjectDir}${_path2.default.sep}`)) {
        throw new Error(`File \`${resolvedProjectDir}\` must be located within project dir \`${projectDir}\`.`);
      }

      // ensure report format is adequate
      if (!['normal', 'json'].includes(report)) {
        throw new Error(`It is not allowed to use \`${report}\` as a value for report. Please use \`normal\` or \`json\`.`);
      }

      // read user module and determine what template to use
      let userModuleContents;
      try {
        userModuleContents = yield (0, _fsExtra.readFile)(userModulePath, 'utf-8');
      } catch (err) {
        throw new Error(`File '${userModulePath}' could not be read.`);
      }
      const argsRegex = new RegExp(`^${outputName} +\\w+ *=`);
      const needArgs = userModuleContents.split('\n').some(function (line) {
        return argsRegex.test(line.trim());
      });

      // read main module template
      _shelljs2.default.cd(resolvedProjectDir);
      let template;
      const chosenTemplatePath = needArgs ? templateWithArgsPath : templatePath;
      try {
        template = yield (0, _fsExtra.readFile)(chosenTemplatePath, 'utf-8');
      } catch (err) {
        throw new Error(`Elm file '${chosenTemplatePath}' does not exist.`);
      }

      // create main module file from template
      const mainModuleContents = template.replace('{mainModule}', mainModule).replace('{userModule}', userModule).replace(/\{output\}/g, outputName);
      yield (0, _fsExtra.writeFile)(mainModuleFilename, mainModuleContents, 'utf-8');

      // compile main module
      // need to fake standard methods due to https://github.com/rtfeldman/node-elm-compiler/issues/68

      var _ref2 = yield Promise.all([(0, _fsExtra.open)(elmCompileStdout, 'w'), (0, _fsExtra.open)(elmCompileStderr, 'w')]);

      var _ref3 = _slicedToArray(_ref2, 2);

      elmCompileStdoutFd = _ref3[0];
      elmCompileStderrFd = _ref3[1];

      const originalProcessExit = process.exit;
      const originalConsoleError = console.error;
      process.exit = noop;
      console.error = noop;
      try {
        yield new Promise(function (resolve) {
          (0, _nodeElmCompiler.compile)([mainModuleFilename], {
            yes: true,
            report,
            pathToMake: pathToElmMake,
            output: outputCompiledFilename,
            processOpts: {
              stdio: [0, elmCompileStdoutFd, elmCompileStderrFd]
            }
          }).on('close', resolve);
        });
      } catch (e) {
        // catching Error: spawn EACCES when path-to-elm-make is a directory
        // (no action needed)
      }
      process.exit = originalProcessExit;
      console.error = originalConsoleError;

      if (!(yield (0, _fsExtra.exists)(outputCompiledFilename))) {
        const errorMessage = `Compilation failed\n${yield (0, _fsExtra.readFile)(elmCompileStderr, 'utf8')}${yield (0, _fsExtra.readFile)(elmCompileStdout, 'utf8')}`;
        throw new Error(errorMessage);
      }

      // run compiled elm file
      const result = {
        output: '',
        debugLog: []
      };
      const originalConsoleLog = console.log;
      console.log = function (...args) {
        return result.debugLog.push(args.join(' '));
      };
      yield new Promise(function (resolve) {
        const worker = require(outputCompiledFilename)[mainModule].worker;

        const app = needArgs ? worker(argsToOutput) : worker();
        app.ports.sendOutput.subscribe(function (output) {
          result.output = output;
          resolve();
        });
      });
      console.log = originalConsoleLog;
      return result;
    } catch (err) {
      // handle error
      let message;
      if (typeof err === 'object' && 'message' in err) {
        if (err.message.indexOf(`does not expose \`${outputName}\``) !== -1) {
          message = `Elm file \`${userModulePath}\` does not define \`${outputName}\`.`;
        } else if (err.message.indexOf(`I cannot find module '${userModule}'`) !== -1) {
          message = `Elm file \`${userModulePath}\` cannot be reached from project dir \`${resolvedProjectDir}\`. Have you configured \`source-directories\` in elm-package.json?`;
        } else {
          message = err.message;
        }
      } else {
        message = err;
      }
      err.message = message && process.platform === 'win32' ? message.replace(/\r\n/g, '\n') : message;
      throw err;
    } finally {
      // cleanup
      yield Promise.all([(0, _fsExtra.close)(elmCompileStdoutFd).catch(noop), (0, _fsExtra.close)(elmCompileStderrFd).catch(noop)]);
      yield Promise.all([(0, _fsExtra.unlink)(mainModuleFilename).catch(noop), (0, _fsExtra.unlink)(outputCompiledFilename).catch(noop), (0, _fsExtra.unlink)(elmCompileStdout).catch(noop), (0, _fsExtra.unlink)(elmCompileStderr).catch(noop)]);
    }
  });

  return function (_x) {
    return _ref.apply(this, arguments);
  };
})();